from chama import simulation
from chama import sensors
from chama import impact
from chama import optimize
from chama import graphics

__version__ = '0.1.2'

__copyright__ = """Copyright 2016-2019 National Technology & Engineering 
    Solutions of Sandia, LLC (NTESS). Under the terms of Contract 
	DE-NA0003525 with NTESS, the U.S. Government retains certain rights 
	in this software."""

__license__ = "Revised BSD License"
